/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.army.entities;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Area;
import java.beans.PropertyChangeSupport;
import java.io.File;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import fb.army.hierarchy.Army;
import fb.army.hierarchy.Command;
import fb.army.hierarchy.HierarchyNode;
import fb.artificialintelligence.AIConstants;
import fb.artificialintelligence.PathFinder;
import fb.battlefield.Sprite;
import fb.battlefield.PaintedWithSprite;
import fb.battlefieldelements.Sign;
import fb.battlefieldelements.Warrior;
import fb.main.FB;
import fb.main.Battle;
import fb.network.Client;
import fb.network.MessagesConstants;
import fb.phases.BinaryPhase;
import fb.phases.MovementPhase;
import fb.phases.PlayerMagicPhase;
import fb.spells.BasicSpell;
import fb.spells.TargetSpell;
import fb.util.FBColors;
import fb.util.FileTools;
import fb.util.GeometryTools;
import fb.util.GraphicTools;
import fb.util.Resources;
import fb.util.StringTools;

/**
 * mother class of all fighting entities (units or individuals)
 * 
 * @author farfpeuf
 * 
 */
public abstract class Entity extends HierarchyNode implements Cloneable,
		PaintedWithSprite {

	// property change constants

	public static final String FLYING = "flying";

	public static final String REMAINING_MOVE = "remainingMove";

	public static final String SCALE = "scale";

	public static final String STATE = "state";

	public static final String LIFEPOINTS = "lifepoints";

	/**
	 * sprite used to paint entity
	 */

	private Sprite sprite;

	/**
	 * entity painting scale (<1 if entity was shrunk by a druid spell)
	 */
	private float scale = 1.0f;

	/**
	 * entity current state
	 */

	private char state = 'n';

	/**
	 * remaining movement points
	 */
	private int remainingMove;

	/**
	 * entity catalog characteristics
	 */
	private int normalStrength, normalResistance, normalWorth, normalMovement;

	private float normalWidth, normalLength;

	protected char type, armor, weapon, distanceWeapon, badLight, genre;

	protected int level, min, max, upgradePrice, lineNumber, initialPdv, price,
			cost, orientation, currentCasualties, initialX, initialY;

	/**
	 * initial orientation angle
	 */

	private double initialAngle = 0;

	/**
	 * sprite id
	 */

	private int spriteNumber;

	/**
	 * number of turn remaining for haste spell
	 */

	private int haste = 0;

	/**
	 * number of turn remaining for magic shield spell
	 */

	private int magicShield = 0;

	/**
	 * number of turn remaining for projectile shield spell
	 */

	private int projectileShield = 0;

	/**
	 * is this entity enabled in the catalog ?
	 */
	private boolean enabled;

	/**
	 * entity portrait image
	 */
	private ImageIcon photo;

	/**
	 * signs and warriors
	 */
	protected List signs, models;

	/**
	 * name and race
	 */
	protected String name, race, fullName;

	/**
	 * command to which this entity belongs
	 */
	protected Command command;

	/**
	 * entity base area and destination area
	 */
	protected Area area, destArea;

	/**
	 * contacts with other entities
	 */
	protected List friendlycontacts, opposingContacts, secondcontacts,
			thirdcontacts;

	protected boolean upgradable,// is this entity upgradable?
			terrible,// is this entity frightening?
			hasShield;// do warriors have shield?

	/**
	 * base vertices
	 */
	protected Point[] vertex;

	/**
	 * signs
	 */
	protected Sign speedSign, magicShieldSign, projectileShieldSign;

	/**
	 * portrait id
	 */

	private int photoIndex;

	/**
	 * create entity from line read in catalog
	 * 
	 * @param ligne
	 *            catalog line
	 * @param linenumber
	 *            line number
	 */
	public Entity(String ligne, int linenumber) {
		this.lineNumber = linenumber;
		if (ligne != null) {
			readCharacteristics(ligne);
		}
	}

	/**
	 * get the army color
	 */
	public Color getColor() {
		if (getArmy() != null) {
			return getArmy().getColor();
		} else {
			return FBColors.standardText;
		}
	}

	/**
	 * reset entity after end of battle
	 * 
	 * @param winner
	 *            did the entity army won the battle
	 */
	public void endBattle(boolean winner) {
		area = null;
		friendlycontacts = null;
		opposingContacts = null;
		secondcontacts = null;
		thirdcontacts = null;
		signs = null;
		models = null;
		magicShieldSign = null;
		speedSign = null;
		projectileShieldSign = null;
		scale = 1.0f;
		magicShield = 0;
		haste = 0;
		projectileShield = 0;
		setX(0);
		setY(0);
		initialAngle = 0;
		initialX = 0;
		initialY = 0;
	}

	/**
	 * initialize entity at the beginning of a battle
	 * 
	 */
	public void initBattle() {
		area = new Area();
		friendlycontacts = new ArrayList();
		opposingContacts = new ArrayList();
		secondcontacts = new ArrayList();
		thirdcontacts = new ArrayList();
		signs = new ArrayList();
		models = new ArrayList();
		magicShieldSign = new Sign(this);
		speedSign = new Sign(this);
		projectileShieldSign = new Sign(this);
		signs.add(speedSign);
		signs.add(magicShieldSign);
		signs.add(projectileShieldSign);
		if (getHaste() > 0) {
			speedSign.setImage(Resources.getSign("speed"));
		}
		if (getMagicShield() > 0) {
			magicShieldSign.setImage(Resources.getSign("magicShield"));
		}
		if (getProjectileShield() > 0) {
			projectileShieldSign
					.setImage(Resources.getSign("projectileShield"));
		}
	}

	/**
	 * read the entity characteristics from catalog line
	 * 
	 * @param line
	 *            the catalog line
	 */
	private void readCharacteristics(String line) {
		StringBuffer buffy = new StringBuffer(line);
		try {
			spriteNumber = StringTools.readBufferInt(buffy);
			setPhoto(StringTools.readBufferInt(buffy));
			race = StringTools.readBuffer(buffy);
			name = StringTools.readBuffer(buffy);
			normalWidth = StringTools.readBufferFloat(buffy);
			normalLength = StringTools.readBufferFloat(buffy);
			level = StringTools.readBufferInt(buffy);
			type = StringTools.readBufferChar(buffy);
			normalStrength = StringTools.readBufferInt(buffy);
			normalResistance = StringTools.readBufferInt(buffy);
			normalWorth = StringTools.readBufferInt(buffy);
			armor = StringTools.readBufferChar(buffy);
			hasShield = StringTools.readBuffer(buffy).equals("yes");
			weapon = StringTools.readBufferChar(buffy);
			distanceWeapon = StringTools.readBufferChar(buffy);
			normalMovement = StringTools.readBufferInt(buffy);
			initialPdv = StringTools.readBufferInt(buffy);
			terrible =StringTools.readBuffer(buffy).equals("yes");
			badLight = StringTools.readBufferChar(buffy);
			String pricestring = StringTools.readBuffer(buffy);
			if (!pricestring.equals("0")) {
				price = Integer.parseInt(pricestring);
			}
			String upgradepricestring = StringTools.readBuffer(buffy);
			if (!upgradepricestring.equals("0")) {
				upgradePrice = Integer.parseInt(upgradepricestring);
				upgradable = true;
			}else{
				upgradable = false;
			}
			String minstring = StringTools.readBuffer(buffy);
			if (!minstring.equals("0")) {
				min = Integer.parseInt(minstring);
			}
			String maxstring = StringTools.readBuffer(buffy);
			if (!maxstring.equals("0")) {
				max = Integer.parseInt(maxstring);
			}
			enabled =StringTools.readBuffer(buffy).equals("yes");
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println("Erreur lecture entit ligne " + lineNumber
					+ " : " + name + " " + race);
		}
	}



	/**
	 * internationalize entity name
	 */
	public void internationalize() {
		String tmpname, arrangedname, arrangedrace;
		tmpname = Resources.getString(name);
		genre = tmpname.charAt(0);
		arrangedname = tmpname.substring(1);
		arrangedrace = Resources.getString(genre + race);
		if (name.indexOf(race) != -1) {
			arrangedrace = "";
		}

		fullName = StringTools.replaceAll(StringTools.replaceAll(Resources
				.getString("fullName"), "<race>", arrangedrace), "<name>",
				arrangedname);
	}

	/**
	 * clone this entity
	 */
	public Object clone() {
		try {
			Entity clone = (Entity) super.clone();
			clone.support = new PropertyChangeSupport(clone);
			return clone;
		} catch (CloneNotSupportedException cnse) {
			cnse.printStackTrace();
			return null;
		}
	}

	/**
	 * set entity portrait
	 * 
	 * @param index
	 *            portrait id
	 */
	public void setPhoto(int index) {
		photoIndex = index;
		String path;
		String imagepath = "portrait" + File.separator + index + ".jpg";
		String defaultpath = "portrait" + File.separator + "default.gif";
		if (FileTools.exists(imagepath)) {
			path = imagepath;
		} else {
			path = defaultpath;
		}
		photo = Resources.getImage(path);
	}

	/**
	 * test if another entity is an ennemy of this entity
	 * 
	 * @param e
	 *            the other entity
	 * @return true if the 2 entities are ennemies
	 */
	public boolean isEnnemy(Entity e) {
		return getArmy() == null
				|| getArmy().getEnnemies().contains(e.getArmy());
	}

	/**
	 * test if this entity is active
	 * 
	 * @return true if the entity wass neither annihilated nor has routed
	 */
	public boolean isActive() {
		return (state != 'r' && state != 'a');
	}

	/**
	 * calculate angle to another entity
	 * 
	 * @param entity
	 *            the other entity
	 * @return the angle
	 */
	public double angleTo(Entity entity) {
		return GeometryTools.getRoundedAngle(getX(), getY(), entity.getX(),
				entity.getY());
	}

	/**
	 * test if line of sight is clear
	 * 
	 * @param losP0
	 *            line of sight first point
	 * @param losP1
	 *            line of sight second point
	 * @param target
	 *            target entity
	 * @return true if line of sight is clear
	 */
	public boolean clearLineOfSight(Point losP0, Point losP1, Entity target) {
		Area obstacles = getLineOfSightObstacles();
		if (target != null) {
			obstacles.subtract(target.area);
		}
		return !GeometryTools.hasIntersection(losP0, losP1, obstacles);
	}

	/**
	 * test if a line has an intersection with obstacles
	 * 
	 * @param line
	 *            the line
	 * @param obstacles
	 *            the obstacles
	 * @return true if the line has an intersection with obstacles
	 */
	public boolean lineClear(Shape line, Area obstacles) {
		Area lineArea = new Area(line);
		lineArea.intersect(obstacles);
		return lineArea.isEmpty();
	}

	/**
	 * test if line of movement is clear
	 * 
	 * @param losP0
	 *            line of movement first point
	 * @param losP1
	 *            line of movement second point
	 * @param target
	 *            target entity
	 * @return true if line of movement is clear
	 */
	public boolean clearLineOfMovement(Point losP0, Point losP1, Entity target) {
		Area obstacles = getMovementObstacles();
		if (target != null) {
			obstacles.subtract(target.area);
		}
		return !GeometryTools.hasIntersection(losP0, losP1, obstacles);
	}

	/**
	 * test if line of movement is clear
	 * 
	 * @param lom
	 *            the line of movement
	 * @param target
	 *            target entity
	 * @return true if line of movement is clear
	 */
	public boolean clearLineOfMovement(Shape lom, Entity target) {
		Area obstacles = getMovementObstacles();
		if (target != null) {
			obstacles.subtract(target.area);
		}
		return lineClear(lom, obstacles);
	}

	/**
	 * get obstacles to entity movement
	 * 
	 * @return the obstacles area
	 */
	public Area getMovementObstacles() {
		return getMovementObstacles(!isFlying());
	}

	/**
	 * get obstacles to entity movement
	 * 
	 * @param groundObstacles
	 *            obstacles on ground or in the sky ?
	 * @return the obstacles area
	 */
	public Area getMovementObstacles(boolean groundObstacles) {
		Area obstacles = new Area();
		if (groundObstacles) {
			obstacles.add(Battle.getInstance().getLineOfMovementObstacles());
		} else {
			obstacles.add(Battle.getInstance().getExterior());
		}
		List entities = Battle.getInstance().getActiveEntities();
		Entity entity;
		for (int i = 0; i < entities.size(); i++) {
			entity = (Entity) entities.get(i);
			if (entity != this && (entity.isFlying() != groundObstacles)
					&& entity.isDeployed()) {
				obstacles.add(entity.area);
			}
		}
		return obstacles;
	}

	/**
	 * get obstacles to entity sight
	 * 
	 * @return the obstacles area
	 */
	public Area getLineOfSightObstacles() {
		return getLineOfSightObstacles(!isFlying());
	}

	/**
	 * get bounds of movement range shape
	 * 
	 * @return smallest rectangle containing movement range shape
	 */
	public Rectangle getMovementRangeBounds() {
		return getMovementRangeBounds(getRemainingMove());
	}

	/**
	 * get bounds of movement range shap
	 * 
	 * @param remainingMovement
	 *            entity remaining movement points
	 * @return smallest rectangle containing movement range shape
	 */
	public Rectangle getMovementRangeBounds(int remainingMovement) {
		int rx = (int) (getX() - remainingMovement * 8);
		int ry = (int) (getY() - remainingMovement * 8);
		int rw = (int) (remainingMovement * 16);
		int rh = (int) (remainingMovement * 16);
		return new Rectangle(rx, ry, rw, rh);
	}

	/**
	 * get bounds of destination shape
	 * 
	 * @return smallest rectangle containing destination shape
	 */
	public Rectangle getDestinationBounds() {
		int rayon = getRemainingMove() + getMaxDiag();

		int rx = (int) (getX() - rayon * 8);
		int ry = (int) (getY() - rayon * 8);
		int rw = (int) (rayon * 16);
		int rh = (int) (rayon * 16);
		return new Rectangle(rx, ry, rw, rh);
	}

	/**
	 * get shape representing movement range
	 * 
	 * @return the shape representing movement range
	 */
	public Area getMovementRange() {
		int mvt = getRemainingMove() * 8;
		int x = getX();
		int y = getY();
		return new Area(new Arc2D.Double(x - mvt, y - mvt, 2 * mvt, 2 * mvt, 0,
				360, Arc2D.PIE));
	}

	/**
	 * get obstacles to entity line of sight
	 * 
	 * @param groundObstacles
	 *            obstacles on ground or in the sky ?
	 * @return the obstacles area
	 */
	public Area getLineOfSightObstacles(boolean groundObstacles) {
		Area obstacles = new Area();
		if (groundObstacles) {
			obstacles.add(Battle.getInstance().getLineOfSightObstacles());
		}
		List activeEntities = Battle.getInstance().getActiveEntities();
		Entity entity;
		for (int i = 0; i < activeEntities.size(); i++) {
			entity = (Entity) activeEntities.get(i);
			if (entity != this && (entity.isFlying() != groundObstacles)) {
				obstacles.add(entity.area);
			}
		}
		return obstacles;
	}

	/**
	 * can this entity move to this target
	 * 
	 * @param target
	 *            the target
	 * @return true if this entity move to its target
	 */
	public boolean moveClear(Entity target) {
		if (canMoveTo(target.getX(), target.getY(), target)) {
			return true;
		}
		Warrior model;
		for (int i = 0; i < target.getModels().size(); i++) {
			model = (Warrior) target.getModels().get(i);
			if (canMoveTo((int) model.getX(), (int) model.getY(), target)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * can this entity reach its target this turn
	 * 
	 * @param target
	 *            the target
	 * @param obstacles
	 *            movement obstacles
	 * @return true if this entity reach its target this turn
	 */
	public boolean canReach(Entity target, Area obstacles) {
		return canMoveTo(target.getX(), target.getY(), obstacles)
				&& (stepsTo(target) <= remainingMove);
	}

	/**
	 * test if this entity can move to specified destination
	 * 
	 * @param destx
	 *            destination x coordinate
	 * @param desty
	 *            destination y coordinate
	 * @param target
	 *            target entity
	 * @return true if this entity can move to specified destination
	 */
	public boolean canMoveTo(int destx, int desty, Entity target) {
		return clearLineOfMovement(getPathTo(destx, desty), target);
	}

	/**
	 * test if this entity can move to specified destination
	 * 
	 * @param destx
	 *            destination x coordinate
	 * @param desty
	 *            destination y coordinate
	 * @param obstacles
	 *            movement obstacles
	 * @return true if this entity can move to specified destination
	 */
	public boolean canMoveTo(int destx, int desty, Area obstacles) {
		return lineClear(getPathTo(destx, desty), obstacles);
	}

	/**
	 * get path area to specified destination
	 * 
	 * @param destx
	 *            destination x coordinate
	 * @param desty
	 *            destination y coordinate
	 * @return path area
	 */
	public Area getPathTo(int destx, int desty) {
		int x, y, dx, dy;
		int[] areax, areay;
		areax = new int[4];
		areay = new int[4];
		Point v, vt;
		Area tarea, parcours;
		x = getX();
		y = getY();
		dx = destx - x;
		dy = desty - y;
		parcours = new Area();
		tarea = new Area(AffineTransform.getTranslateInstance(dx, dy)
				.createTransformedShape(area));
		parcours.add(tarea);
		areax[0] = x;
		areax[1] = destx;
		areay[0] = y;
		areay[1] = desty;
		for (int i = 0; i < vertex.length; i++) {
			v = vertex[i];
			vt = new Point(v);
			vt.translate(dx, dy);
			areax[2] = (int) (vt.getX());
			areax[3] = (int) (v.getX());
			areay[2] = (int) (vt.getY());
			areay[3] = (int) (v.getY());
			parcours.add(new Area(new Polygon(areax, areay, 4)));
		}
		return parcours;
	}

	/**
	 * try to move entity to target entity
	 * 
	 * @param e
	 *            target entity
	 */
	public void moveTo(Entity e) {
		moveTo(e, e.getX(), e.getY());
	}

	/**
	 * try to move entity to target entity
	 * 
	 * @param e
	 *            target entity
	 * @param ex
	 *            destination x coordinate
	 * @param ey
	 *            destination y coordinate
	 * @return true if a path was found
	 */
	public boolean moveTo(final Entity e, final int ex, final int ey) {
		boolean pathFound;
		int xdest, ydest;
		PathFinder pathFinder = new PathFinder(this, e, ex, ey);
		synchronized (this) {
			pathFinder.start();
			try {
				wait(FB.getInstance().isBatch() ? 200
						: AIConstants.MAX_PATH_SEARCHING_TIME);
			} catch (InterruptedException e1) {
				e1.printStackTrace();
			}
		}
		pathFound = pathFinder.isPathFound();
		xdest = pathFinder.getDestX();
		ydest = pathFinder.getDestY();
		if (pathFinder.isAlive()) {
			PathFinder moribund = pathFinder;
			pathFinder = null;
			moribund.safeStop();
		}
		if (pathFound) {
			moveto(xdest, ydest);
		} else {
			// FB.debug("no path found");
		}
		return pathFound;
	}

	/**
	 * paint entity border
	 * 
	 * @param g
	 *            graphics
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param scale
	 *            view scale
	 */
	public void paintBorder(Graphics g, int dx, int dy, int scale) {
		if (isActive() && !isWalking()) {
			Graphics2D g2d = (Graphics2D) g;
			Area scalecontour = new Area(AffineTransform.getScaleInstance(
					scale, scale).createTransformedShape(getBorder()));
			scalecontour = new Area(AffineTransform
					.getTranslateInstance(dx, dy).createTransformedShape(
							scalecontour));
			g2d.draw(scalecontour);
		}
	}

	/**
	 * paint shape for a possible movement
	 * 
	 * @param g
	 *            graphics
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param mouseX
	 *            mouse x coordinate
	 * @param mouseY
	 *            mouse y coordinate
	 */
	public void paintMovementOK(Graphics g, int dx, int dy, int mouseX,
			int mouseY) {
		g.setColor(FBColors.secondary);
		int movelength = (int) (0.99 + Math.sqrt(dx * dx + dy * dy) / 8);
		g.drawString(Integer.toString(movelength), mouseX, mouseY);
	}

	/**
	 * get entity tool type
	 * 
	 * @param phase
	 *            current game phase
	 * @return entity tool tip
	 */
	public String getToolTip(BinaryPhase phase) {
		if (phase != null && phase instanceof PlayerMagicPhase) {
			BasicSpell spell = ((PlayerMagicPhase) phase).getSpell();
			if (!phase.isActorSelection() && spell instanceof TargetSpell
					&& !spell.getCaster().isCastingSpell()
					&& phase.getPotentialTargets().contains(this)) {
				StringBuffer statString = new StringBuffer("<html>");
				statString.append(Resources.getString("averagecost"));
				statString.append(((TargetSpell) spell).averageCost(spell
						.getCaster(), this));
				statString.append("<br>");
				statString.append(Resources.getString("mincost"));
				statString.append(((TargetSpell) spell).minCost(spell
						.getCaster(), this));
				statString.append("<br>");
				statString.append(Resources.getString("maxcost"));
				statString.append(((TargetSpell) spell).maxCost(spell
						.getCaster(), this));
				statString.append("</html>");
				return statString.toString();
			}
		}
		return null;
	}

	/**
	 * get bounds of deployment shape
	 * 
	 * @param tX
	 *            deployment destination x coordinate
	 * @param tY
	 *            deployment destination y coordinate
	 * @return smallest rectangle containing destination shape
	 */
	public Rectangle getDeploymentBounds(int tX, int tY) {
		return getDeploymentArea(tX, tY).getBounds();
	}

	/**
	 * get deployment area
	 * 
	 * @param tX
	 *            deployment destination x coordinate
	 * @param tY
	 *            deployment destination y coordinate
	 * @return deployment area
	 */
	public Area getDeploymentArea(int tX, int tY) {
		int mdx = tX - getX();
		int mdy = tY - getY();
		return new Area(AffineTransform.getTranslateInstance(mdx, mdy)
				.createTransformedShape(area));

	}

	/**
	 * paint deployment area
	 * 
	 * @param g
	 *            graphics
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param scale
	 *            view scale
	 * @param deployArea
	 *            deployment area
	 * @param mouseX
	 *            mouse x coordinate
	 * @param mouseY
	 *            mouse y coordinate
	 */
	public void paintDeployment(Graphics g, int dx, int dy, int scale,
			Area deployArea, int mouseX, int mouseY) {
		int tX, tY;
		Area scaletarea;
		tX = (mouseX - dx) / scale;
		tY = (mouseY - dy) / scale;

		if (canDeployTo(tX, tY, deployArea)) {
			g.setColor(FBColors.secondary);
			FB.getInstance().setBattlefieldCursor(FB.DEFAULT_CURSOR);
		} else {
			g.setColor(FBColors.errorColor);
			FB.getInstance().setBattlefieldCursor(FB.NO_CURSOR);
		}
		scaletarea = new Area(AffineTransform.getScaleInstance(scale, scale)
				.createTransformedShape(getDeploymentArea(tX, tY)));
		scaletarea = new Area(AffineTransform.getTranslateInstance(dx, dy)
				.createTransformedShape(scaletarea));
		((Graphics2D) g).draw(scaletarea);
	}

	/**
	 * paint movement shape
	 * 
	 * @param g
	 *            graphics
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param scale
	 *            view scale
	 * @param mouseX
	 *            mouse x coordinate
	 * @param mouseY
	 *            mouse y coordinate
	 * @param okMove
	 *            is movement possible
	 */
	public void paintMovement(Graphics g, int dx, int dy, int scale,
			int mouseX, int mouseY, int okMove) {
		int destx, desty, distx, disty;

		Shape destShape;
		destx = (mouseX - dx) / scale;
		desty = (mouseY - dy) / scale;
		distx = destx - getX();
		disty = desty - getY();

		int dist = (int) java.lang.Math.sqrt(distx * distx + disty * disty) / 8;

		if (okMove != MovementPhase.MOVE_TO_TARGET && dist < getRemainingMove()) {
			if (okMove == MovementPhase.MOVE_TO_PLACE) {
				paintMovementOK(g, distx, disty, mouseX, mouseY);
			} else {
				g.setColor(FBColors.errorColor);
				Point v;
				int vx, vy;
				for (int i = 0; i < vertex.length; i++) {
					v = vertex[i];
					vx = (int) v.getX();
					vy = (int) v.getY();
					g.drawLine(vx * scale + dx, vy * scale + dy, (vx + distx)
							* scale + dx, (vy + disty) * scale + dy);
				}
				g.setColor(FBColors.errorColor);
			}
			destShape = AffineTransform.getTranslateInstance(distx, disty)
					.createTransformedShape(area);
			destShape = GraphicTools.getScaledShape(destShape, dx, dy, scale);
			((Graphics2D) g).draw(destShape);
		}
	}

	/**
	 * has this entity lost life points
	 * 
	 * @return
	 */
	public boolean hasLostLifePoints() {
		return getTotalLifePoints() < getInitialTotalLifePoints();
	}

	/**
	 * get warriors
	 * 
	 * @return the warriors
	 */
	public List getModels() {
		return models;
	}

	/**
	 * get warrior by index
	 * 
	 * @param index
	 *            model index
	 * @return the warrior
	 */
	public Warrior getModel(int index) {
		return (Warrior) models.get(index);
	}

	/**
	 * get number of steps to another entity
	 * 
	 * @param e
	 *            the target entity
	 * @return number of steps
	 */
	public int stepsTo(Entity e) {
		return distanceTo(e) / 8;
	}

	/**
	 * move entity to specified destination
	 * 
	 * @param x
	 *            destination x coordinate
	 * @param y
	 *            destination y coordinate
	 */
	public void moveto(int x, int y) {
		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.MOVETO + getId() + "#" + x + "#" + y);
		} else {
			doMoveTo(x, y);
		}
	}

	/**
	 * deploy entity to specified destination
	 * 
	 * @param x
	 *            destination x coordinate
	 * @param y
	 *            destination y coordinate
	 */
	public void deployTo(int x, int y) {
		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.DEPLOYTO + getId() + "#" + x + "#" + y);
		} else {
			doDeployTo(x, y);
		}
	}

	/**
	 * do deploy entity to specified destination
	 * 
	 * @param x
	 *            destination x coordinate
	 * @param y
	 *            destination y coordinate
	 */
	public void doDeployTo(int x, int y) {
		doMoveTo(x, y, false);
		teleport();
	}

	/**
	 * rotate entity
	 * 
	 * @param theta
	 *            angle
	 */
	public void rotate(double theta) {
		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.ROTATE + getId() + "#" + theta);
		} else {
			doRotate(theta);
		}
	}

	/**
	 * do move entity to specified destination
	 * 
	 * @param x
	 *            destination x coordinate
	 * @param y
	 *            destination y coordinate
	 */
	public void doMoveTo(int x, int y) {
		doMoveTo(x, y, true);
	}

	/**
	 * d100 score needed to kill target
	 * 
	 * @param target
	 *            target entity
	 * @param weapon
	 *            weapon used
	 * @return score needed to kill target
	 */
	public int rollToKill(Entity target, char weapon) {
		if (weapon == '1') {
			if (target.armor == 'x') {
				return 80;
			} else if (target.armor == 'h') {
				return 80;
			} else if (target.armor == 'm') {
				return 65;
			} else if (target.armor == 'l') {
				return 50;
			} else {
				return 35;
			}
		} else if (weapon == '2') {
			if (target.armor == 'x') {
				return 65;
			} else if (target.armor == 'h') {
				return 50;
			} else if (target.armor == 'm') {
				return 35;
			} else {
				return 20;
			}
		} else if (weapon == 'h') {
			if (target.armor == 'x') {
				return 80;
			} else if (target.armor == 'h') {
				return 50;
			} else {
				return 35;
			}
		} else if (weapon == 't') {
			if (target.armor == 'x') {
				return 80;
			} else if (target.armor == 'h') {
				return 65;
			} else if (target.armor == 'm') {
				return 50;
			} else if (target.armor == 'l') {
				return 35;
			} else {
				return 20;
			}
		} else {
			if (target.armor == 'x') {
				return 80;
			} else if (target.armor == 'h') {
				return 80;
			} else if (target.armor == 'm') {
				return 65;
			} else if (target.armor == 'l') {
				return 50;
			} else {
				return 35;
			}
		}
	}

	/**
	 * create entity warriors and calculate base area
	 * 
	 */
	public void create() {
		createModel();
		calculateArea();
	}

	/**
	 * upgrade entity level
	 * 
	 */
	public void upgradeLevel() {
		level++;
	}

	/**
	 * downgrade entity level
	 * 
	 */
	public void downgradeLevel() {
		level--;
	}

	/**
	 * calculate line of sight length to another entity
	 * 
	 * @param e
	 *            the target entity
	 * @return line of sight length or Integer.MAX_VALUE if line of sight is not
	 *         clear
	 */
	public int lineOfSightTo(Entity e) {
		boolean losClear = clearLineOfSight(new Point(getX(), getY()),
				new Point(e.getX(), e.getY()), e);
		if (losClear) {
			int X = e.getX() - getX();
			int Y = e.getY() - getY();
			return (int) Math.sqrt(X * X + Y * Y);
		}

		return Integer.MAX_VALUE;

	}

	/**
	 * is this entity currently suffering from bad light
	 * 
	 * @return true if this entity currently suffering from bad light
	 */
	public boolean suffersBadLight() {
		return (Battle.getInstance().getTime() <= 5 && badLight == 'd')
				|| (Battle.getInstance().getTime() > 5 && badLight == 'n');
	}

	/**
	 * get entity level as a string
	 * 
	 * @return entity level as a string
	 */
	public String levelString() {
		return Integer.toString(level);
	}

	/**
	 * get entity type as a string
	 * 
	 * @return entity type as a string
	 */
	public String typeString() {
		if (type == 'c') {
			return Resources.getString("Calm");
		} else if (type == 'a') {
			return Resources.getString("Agitated");
		} else if (type == 'f') {
			return Resources.getString("Fanatic");
		} else if (type == 'u') {
			return Resources.getString("Undead");
		} else
			return "?";
	}

	/**
	 * get entity armor as a string
	 * 
	 * @return entity armor as a string
	 */
	public String armorString() {
		if (armor == 'x') {
			return Resources.getString("Xtra");
		} else if (armor == 'h') {
			return Resources.getString("Heavy");
		} else if (armor == 'm') {
			return Resources.getString("Medium");
		} else if (armor == 'l') {
			return Resources.getString("Light");
		} else
			return Resources.getString("None");
	}

	/**
	 * get entity weapon as a string
	 * 
	 * @return entity weapon as a string
	 */
	public String weaponString() {
		if (weapon == '1') {
			return Resources.getString("1Hand");
		} else if (weapon == '2') {
			return Resources.getString("2Hand");
		} else if (weapon == 't') {
			return Resources.getString("T&C");
		} else if (weapon == 'h') {
			return Resources.getString("H&H");
		} else if (weapon == 'p') {
			return Resources.getString("Pike");
		} else if (weapon == 's') {
			return Resources.getString("Spear");
		} else
			return "-";
	}

	/**
	 * get entity strength as a string
	 * 
	 * @return entity strength as a string
	 */
	public String strengthString() {
		return Integer.toString(getStrength());
	}

	/**
	 * get entity resistance as a string
	 * 
	 * @return entity resistance as a string
	 */
	public String resistanceString() {
		return Integer.toString(getResistance());
	}

	/**
	 * get entity worth as a string
	 * 
	 * @return entity worth as a string
	 */
	public String worthString() {
		return Integer.toString(getWorth());
	}

	/**
	 * get entity movement as a string
	 * 
	 * @return entity movement as a string
	 */
	public String movementString() {
		return Integer.toString(getMovement());
	}

	/**
	 * get entity life points as a string
	 * 
	 * @return entity life points as a string
	 */
	public String lifePointsString() {
		return Integer.toString(getTotalLifePoints());
	}

	/**
	 * get entity price as a string
	 * 
	 * @return entity price as a string
	 */
	public String priceString() {
		return Integer.toString(getPrice());
	}

	/**
	 * add a contact with another entity
	 * 
	 * @param entity2
	 *            the other entity
	 * @param nature
	 *            type of contact
	 */
	public void addContact(Entity entity2, int nature) {
		if (nature == 0) {
			friendlycontacts.add(entity2);
		} else if (nature == 1) {
			opposingContacts.add(entity2);
		} else if (nature == 2) {
			secondcontacts.add(entity2);
		} else {
			thirdcontacts.add(entity2);
		}
	}

	/**
	 * remove a contact with another entity
	 * 
	 * @param entity2
	 *            the other entity
	 * @param nature
	 *            type of contact
	 */
	public void removeContact(Entity entity2, int nature) {
		if (nature == 0) {
			friendlycontacts.remove(entity2);
		} else if (nature == 1) {
			opposingContacts.remove(entity2);
		} else if (nature == 2) {
			secondcontacts.remove(entity2);
		} else {
			thirdcontacts.remove(entity2);
		}
	}

	/**
	 * is this entity near its warchief
	 * 
	 * @return true if this entity is near its warchief
	 */
	public boolean nearWarchief() {
		return getArmy() != null && getArmy().getWarchief() != null
				&& near(getArmy().getWarchief());
	}

	/**
	 * is this entity near another entity
	 * 
	 * @param e
	 *            the other entity
	 * @return true if this entity is near the other entity
	 */
	public boolean near(Entity e) {
		if (e.isActive() && stepsTo(e) < 5) {
			return true;
		}
		return false;
	}

	/**
	 * can this entity move
	 * 
	 * @return true if this entity can move
	 */
	public boolean isMoveAble() {
		return opposingContacts.size() == 0;
	}

	/**
	 * get number of frames in entity sprite
	 */
	public int getNumberOfFrames() {
		return 120;
	}

	/**
	 * add sequences to sprite
	 */
	public void addSpriteSequences(Sprite sprite) {
		sprite.addSpritsequence("stand0", 28, 28, 28, 28);
		sprite.addSpritsequence("stand1", 29, 29, 29, 29);
		sprite.addSpritsequence("stand2", 30, 30, 30, 30);
		sprite.addSpritsequence("stand3", 31, 31, 31, 31);
		sprite.addSpritsequence("stand4", 24, 24, 24, 24);
		sprite.addSpritsequence("stand5", 25, 25, 25, 25);
		sprite.addSpritsequence("stand6", 26, 26, 26, 26);
		sprite.addSpritsequence("stand7", 27, 27, 27, 27);
		sprite.addSpritsequence("fight0", 92, 100, 108, 116);
		sprite.addSpritsequence("fight1", 93, 101, 109, 117);
		sprite.addSpritsequence("fight2", 94, 102, 110, 118);
		sprite.addSpritsequence("fight3", 95, 103, 111, 119);
		sprite.addSpritsequence("fight4", 88, 96, 104, 112);
		sprite.addSpritsequence("fight5", 89, 97, 105, 113);
		sprite.addSpritsequence("fight6", 90, 98, 106, 114);
		sprite.addSpritsequence("fight7", 91, 99, 107, 115);
		sprite.addSpritsequence("walk0", 60, 68, 76, 84);
		sprite.addSpritsequence("walk1", 61, 69, 77, 85);
		sprite.addSpritsequence("walk2", 62, 70, 78, 86);
		sprite.addSpritsequence("walk3", 63, 71, 79, 87);
		sprite.addSpritsequence("walk4", 56, 64, 72, 80);
		sprite.addSpritsequence("walk5", 57, 65, 73, 81);
		sprite.addSpritsequence("walk6", 58, 66, 74, 82);
		sprite.addSpritsequence("walk7", 59, 67, 75, 83);
		sprite.addSpritsequence("die0", 4, 12, 20, 20);
		sprite.addSpritsequence("die1", 5, 13, 21, 21);
		sprite.addSpritsequence("die2", 6, 14, 22, 22);
		sprite.addSpritsequence("die3", 7, 15, 23, 23);
		sprite.addSpritsequence("die4", 0, 8, 16, 16);
		sprite.addSpritsequence("die5", 1, 9, 17, 17);
		sprite.addSpritsequence("die6", 2, 10, 18, 18);
		sprite.addSpritsequence("die7", 3, 11, 19, 19);
		sprite.addSpritsequence("threat0", 36, 44, 52, 44);
		sprite.addSpritsequence("threat1", 37, 45, 53, 45);
		sprite.addSpritsequence("threat2", 38, 46, 54, 46);
		sprite.addSpritsequence("threat3", 39, 47, 55, 47);
		sprite.addSpritsequence("threat4", 32, 40, 48, 40);
		sprite.addSpritsequence("threat5", 33, 41, 49, 41);
		sprite.addSpritsequence("threat6", 34, 42, 50, 42);
		sprite.addSpritsequence("threat7", 35, 43, 51, 43);

	}

	/**
	 * create entity sprite
	 * 
	 */
	public void createSprite() {
		sprite = Sprite.getInstance(this);
	}

	/**
	 * delete sprite to free memory
	 * 
	 */
	public void freeSprite() {
		sprite = null;
	}

	/**
	 * get entity id
	 * 
	 * @return entity id
	 */
	public int getId() {
		return Battle.getInstance().getAllEntities().indexOf(this);
	}

	/**
	 * Returns the banner.
	 * 
	 * @return boolean
	 */
	public boolean isBanner() {
		return this instanceof Banner;
	}

	/**
	 * Returns the courier.
	 * 
	 * @return boolean
	 */
	public boolean isCourier() {
		return this instanceof Courier;
	}

	/**
	 * Returns the crew.
	 * 
	 * @return boolean
	 */
	public boolean isCrew() {
		return this instanceof Crew;
	}

	/**
	 * Returns the destroyer.
	 * 
	 * @return boolean
	 */
	public boolean isDestroyer() {
		return this instanceof Exterminator;
	}

	/**
	 * Returns the dragon.
	 * 
	 * @return boolean
	 */
	public boolean isDragon() {
		return this instanceof Dragon;
	}

	/**
	 * Returns the druid.
	 * 
	 * @return boolean
	 */
	public boolean isDruid() {
		return this instanceof Druid;
	}

	/**
	 * Returns the flyingshootingunit.
	 * 
	 * @return boolean
	 */
	public boolean isFlyingshootingunit() {
		return this instanceof FlyingShootingUnit;
	}

	/**
	 * Returns the flyingunit.
	 * 
	 * @return boolean
	 */
	public boolean isFlyingunit() {
		return this instanceof FlyingUnit;
	}

	/**
	 * Returns the hero.
	 * 
	 * @return boolean
	 */
	public boolean isHero() {
		return this instanceof Hero;
	}

	/**
	 * Returns the individual.
	 * 
	 * @return boolean
	 */
	public boolean isIndividual() {
		return this instanceof Individual;
	}

	/**
	 * Returns the magic.
	 * 
	 * @return boolean
	 */
	public boolean isMagic() {
		return this instanceof SpellCaster;
	}

	/**
	 * Returns the marksman.
	 * 
	 * @return boolean
	 */
	public boolean isMarksman() {
		return this instanceof Marksman;
	}

	/**
	 * Returns the messenger.
	 * 
	 * @return boolean
	 */
	public boolean isMessageReceiver() {
		return this instanceof MessageReceiver;
	}


	/**
	 * Returns the leader.
	 * 
	 * @return boolean
	 */
	public boolean isLeader() {
		return this instanceof Leader;
	}

	/**
	 * Returns the paladin.
	 * 
	 * @return boolean
	 */
	public boolean isPaladin() {
		return this instanceof Legend;
	}

	/**
	 * Returns the priest.
	 * 
	 * @return boolean
	 */
	public boolean isPriest() {
		return this instanceof Priest;
	}

	/**
	 * Returns the scout.
	 * 
	 * @return boolean
	 */
	public boolean isScout() {
		return this instanceof Scout;
	}

	/**
	 * Returns the shootingunit.
	 * 
	 * @return boolean
	 */
	public boolean isShootingUnit() {
		return this instanceof Shootingunit;
	}


	/**
	 * Returns the unit.
	 * 
	 * @return boolean
	 */
	public boolean isUnit() {
		return this instanceof Unit;
	}

	/**
	 * Returns the wizard.
	 * 
	 * @return boolean
	 */
	public boolean isWizard() {
		return this instanceof Wizard;
	}


	/**
	 * Returns the state.
	 * 
	 * @return char
	 */
	public char getState() {
		return state;
	}

	/**
	 * Sets the state.
	 * 
	 * @param state
	 *            The state to set
	 */
	public void setState(char state) {
		char oldValue = this.state;
		this.state = state;
		firePropertyChange(STATE, oldValue, state);
	}

	/**
	 * Returns the currentCasualties.
	 * 
	 * @return int
	 */
	public int getCurrentCasualties() {
		return currentCasualties;
	}

	/**
	 * Returns the remainingMove.
	 * 
	 * @return int
	 */
	public int getRemainingMove() {
		return remainingMove;
	}

	/**
	 * Sets the currentCasualties.
	 * 
	 * @param currentCasualties
	 *            The currentCasualties to set
	 */
	public void setCurrentcasualties(int currentcasualties) {
		this.currentCasualties = currentcasualties;
	}

	/**
	 * reset remaining movement points
	 * 
	 */
	public void resetRemainingMove() {
		if (isMoveAble()) {
			setRemainingmove(getMovement());
		} else {
			setRemainingmove(2);
		}
	}

	/**
	 * Sets the remainingMove.
	 * 
	 * @param remainingMove
	 *            The remainingMove to set
	 */
	public void setRemainingmove(int remainingmove) {
		int oldValue = this.remainingMove;
		this.remainingMove = remainingmove;
		firePropertyChange(REMAINING_MOVE, oldValue, remainingmove);
	}

	/**
	 * returns entity as a string to save it to a file
	 * 
	 * @return entity as a string
	 */
	public String saveString() {
		return "e" + '\t' + lineNumber + '\t' + getX() + '\t' + getY() + '\t'
				+ getAngle() + '\t' + getState() + '\t'
				+ getCurrentCasualties() + '\t' + getRemainingMove() + '\t'
				+ getScale() + '\t' + getHaste() + '\t' + getProjectileShield()
				+ '\t' + getMagicShield();
	}

	/**
	 * load entity from a StringBuffer
	 * 
	 * @param buffy
	 *            the StringBuffer
	 */
	public void loadString(StringBuffer buffy) {
		setInitialX(StringTools.readBufferInt(buffy));
		setInitialY(StringTools.readBufferInt(buffy));
		setInitialAngle(StringTools.readBufferDouble(buffy));
		setState(StringTools.readBufferChar(buffy));
		setCurrentcasualties(StringTools.readBufferInt(buffy));
		setRemainingmove(StringTools.readBufferInt(buffy));
		setScale(StringTools.readBufferFloat(buffy));
		setHaste(StringTools.readBufferInt(buffy));
		setProjectileShield(StringTools.readBufferInt(buffy));
		setMagicShield(StringTools.readBufferInt(buffy));
	}

	/**
	 * Sets the initialAngle.
	 * 
	 * @param initialAngle
	 *            The initialAngle to set
	 */
	public void setInitialAngle(double initialAngle) {
		this.initialAngle = initialAngle;
	}

	/**
	 * Returns the initialAngle.
	 * 
	 * @return double
	 */
	public double getInitialAngle() {
		return initialAngle;
	}

	/**
	 * Returns the initialX.
	 * 
	 * @return int
	 */
	public int getInitialX() {
		return initialX;
	}

	/**
	 * Returns the initialY.
	 * 
	 * @return int
	 */
	public int getInitialY() {
		return initialY;
	}

	/**
	 * Sets the initialX.
	 * 
	 * @param initialX
	 *            The initialX to set
	 */
	public void setInitialX(int initialX) {
		this.initialX = initialX;
	}

	/**
	 * Sets the initialY.
	 * 
	 * @param initialY
	 *            The initialY to set
	 */
	public void setInitialY(int initialY) {
		this.initialY = initialY;
	}

	/**
	 * @see javax.swing.tree.TreeNode#children()
	 */
	public Enumeration children() {
		return null;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getAllowsChildren()
	 */
	public boolean getAllowsChildren() {
		return false;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getChildAt(int)
	 */
	public TreeNode getChildAt(int childIndex) {
		return null;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getChildCount()
	 */
	public int getChildCount() {
		return 0;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getIndex(TreeNode)
	 */
	public int getIndex(TreeNode node) {
		return 0;
	}

	/**
	 * @see javax.swing.tree.TreeNode#getParent()
	 */
	public TreeNode getParent() {
		return command;
	}

	/**
	 * @see javax.swing.tree.TreeNode#isLeaf()
	 */
	public boolean isLeaf() {
		return true;
	}

	/**
	 * remove thi entity from its command
	 * 
	 */
	public void removeFromCommand() {
		command.fireEntity(this);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#insert(MutableTreeNode, int)
	 */
	public void insert(MutableTreeNode child, int index) {
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#remove(int)
	 */
	public void remove(int index) {
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#remove(MutableTreeNode)
	 */
	public void remove(MutableTreeNode node) {
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#removeFromParent()
	 */
	public void removeFromParent() {
		command.remove(this);
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#setParent(MutableTreeNode)
	 */
	public void setParent(MutableTreeNode newParent) {
		command = (Command) newParent;
	}

	/**
	 * @see javax.swing.tree.MutableTreeNode#setUserObject(Object)
	 */
	public void setUserObject(Object object) {
	}

	/**
	 * @return Returns the fullName.
	 */
	public String getFullName() {
		return fullName;
	}

	/**
	 * @return Returns the enabled.
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * @param enabled
	 *            The enabled to set.
	 */
	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}

	/**
	 * can this entity be deployed at specified location
	 * 
	 * @param nx
	 *            deployment location x coordinate
	 * @param ny
	 *            deployment location y coordinate
	 * @param deployArea
	 *            deplyment area
	 * @return true if this entity can be deployed at specified location
	 */
	public boolean canDeployTo(int nx, int ny, Area deployArea) {
		int xt = nx - getX(), yt = ny - getY();
		Area testarea = new Area(AffineTransform.getTranslateInstance(
				(double) xt, (double) yt).createTransformedShape(area));
		testarea.subtract(deployArea);
		return testarea.isEmpty();
	}

	/**
	 * get entity mini icon corresponding to entity type
	 */
	public Icon getMinicon() {
		return Resources.getMinicon(getResourceKey());
	}

	/**
	 * get entity type description for tooltip
	 * 
	 * @return entity type description for tooltip
	 */
	public String getTooltip() {
		return Resources.getString(getResourceKey() + ".tt");
	}

	/**
	 * @return Returns the sprite.
	 */
	public Sprite getSprite() {
		return sprite;
	}

	/**
	 * @return Returns the spriteNumber.
	 */
	public int getSpriteID() {
		return spriteNumber;
	}

	/**
	 * is this entity at the same altitude as anothe rentity
	 * 
	 * @param entity
	 *            the other entity
	 * @return true if both entities are at the same altitude
	 */
	public boolean sameAltitude(Entity entity) {
		return isFlying() == entity.isFlying();
	}

	/**
	 * is this entity currently flying
	 * 
	 * @return true if this entity currently flying
	 */
	public boolean isFlying() {
		return false;
	}

	/**
	 * get entity length
	 */
	public float getLength() {
		return normalLength * scale;
	}

	/**
	 * get entity width
	 */
	public float getWidth() {
		return normalWidth * scale;
	}

	/**
	 * @return Returns the scale.
	 */
	public float getScale() {
		return scale;
	}

	/**
	 * change entity scale
	 * 
	 * @param scale
	 *            the new scale
	 */
	public void setScale(float scale) {
		float oldscale = this.scale;
		this.scale = scale;
		firePropertyChange(SCALE, new Float(oldscale), new Float(scale));
	}

	/**
	 * shrink this entity
	 * 
	 */
	public void shrink() {
		float oldscale = scale;

		while (scale * 2.0f > oldscale) {
			try {
				setScale(scale - 0.1f * oldscale);
				Thread.sleep(125);
			} catch (InterruptedException e) {
				FB.exception(e);
			}
		}
		setScale(0.5f * oldscale);

		setRemainingmove((int) (remainingMove * scale));
	}

	/**
	 * get this entity movement points
	 * 
	 * @return this entity movement points
	 */
	public int getMovement() {
		int mouvement = (int) (normalMovement * scale);
		if (haste > 0) {
			mouvement = (int) (mouvement * 1.5 + 0.99);
		}
		if (mouvement < 1) {
			return 1;
		}
		return mouvement;
	}

	/**
	 * get entity resistance
	 * 
	 * @return entity resistance
	 */
	public int getResistance() {
		return (int) (normalResistance * scale) - (scale < 1.0f ? 1 : 0);
	}

	/**
	 * get entity strength
	 * 
	 * @return entity strength
	 */
	public int getStrength() {
		return (int) (normalStrength * scale) - (scale < 1.0f ? 1 : 0);
	}

	/**
	 * get entity worth
	 * 
	 * @return entity worth
	 */
	public int getWorth() {
		return (int) (normalWorth * scale > 1 ? normalWorth * scale : 1);
	}

	/**
	 * betray and change to another army
	 * 
	 * @param ennemyarmy
	 *            the new army
	 */
	public void betray(Army ennemyarmy) {
		getArmy().removeEntity(this);
		ennemyarmy.addEntity(this);
		calculateContacts();
	}

	/**
	 * is this entity handled by this computer
	 * 
	 * @return true if this entity is handled by this compute
	 */
	public boolean isHandled() {
		return getArmy().isHandled();
	}

	/**
	 * @return Returns the haste.
	 */
	public int getHaste() {
		return haste;
	}

	/**
	 * init entity haste
	 * 
	 */
	public void initHaste() {
		if (remainingMove > 0) {
			setRemainingmove(remainingMove + (int) (getMovement() * .5 + .99));
		}
		haste += 3;
		speedSign.setImage(Resources.getSign("speed"));
	}

	/**
	 * consume haste effect
	 * 
	 */
	public void consumeHaste() {
		haste--;
		if (haste <= 0) {
			speedSign.clear();
		}
	}

	public int getMagicShield() {
		return magicShield;
	}

	/**
	 * init entity magic shield
	 * 
	 */
	public void initMagicShield() {
		magicShield += 5;
		magicShieldSign.setImage(Resources.getSign("magicShield"));
	}

	/**
	 * consume magic shield effect
	 * 
	 */
	public void consumeMagicShield() {
		magicShield--;
		if (magicShield <= 0) {
			magicShieldSign.clear();
		}
	}

	/**
	 * init entity projectile shield
	 * 
	 */

	public void initProjectileShield() {
		projectileShield += 3;
		projectileShieldSign.setImage(Resources.getSign("projectileShield"));
	}

	/**
	 * consume magic projectile effect
	 * 
	 */
	public void consumeProjectileShield() {
		projectileShield--;
		if (projectileShield <= 0) {
			projectileShieldSign.clear();
		}
	}

	/**
	 * @return Returns the projectileShield.
	 */
	public int getProjectileShield() {
		return projectileShield;
	}

	/**
	 * @param projectileShield
	 *            The projectileShield to set.
	 */
	public void setProjectileShield(int projectileShield) {
		this.projectileShield = projectileShield;
	}

	/**
	 * @param haste
	 *            The haste to set.
	 */
	public void setHaste(int haste) {
		this.haste = haste;
	}

	/**
	 * @param magicShield
	 *            The magicShield to set.
	 */
	public void setMagicShield(int magicShield) {
		this.magicShield = magicShield;
	}

	/**
	 * consume movement
	 * 
	 * @param minus
	 *            number of movement points to remove
	 */
	public void consumeMovement(int minus) {
		setRemainingmove(remainingMove - minus);
	}

	/**
	 * is this entity deployed on the battlefield
	 * 
	 * @return true if this entity is deployed on the battlefield
	 */
	public boolean isDeployed() {
		return getX() != 0 && getY() != 0;
	}

	/**
	 * get entity full name
	 */
	public String toString() {
		return getFullName();
	}

	/**
	 * @return Returns the destArea.
	 */
	public Area getDestArea() {
		return destArea;
	}

	/**
	 * @return Returns the group.
	 */
	public Command getCommand() {
		return command;
	}

	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return Returns the photo.
	 */
	public ImageIcon getPhoto() {
		return photo;
	}

	/**
	 * @return Returns the race.
	 */
	public String getRace() {
		return race;
	}

	/**
	 * @return Returns the army.
	 */
	public Army getArmy() {
		if (command == null) {
			return null;
		}
		return command.getArmy();
	}

	/**
	 * is this entity a warchief
	 * 
	 * @return true if this entity is a warchief
	 */
	public boolean isWarchief() {
		return getArmy() != null && getArmy().getWarchief() == this;
	}

	/**
	 * @return Returns the area.
	 */
	public Area getArea() {
		return area;
	}

	/**
	 * @return Returns the badLight.
	 */
	public char getBadLight() {
		return badLight;
	}

	/**
	 * @return Returns the hasShield.
	 */
	public boolean hasShield() {
		return hasShield;
	}

	/**
	 * @return Returns the initialPdv.
	 */
	public int getInitialPdv() {
		return initialPdv;
	}

	/**
	 * @return Returns the level.
	 */
	public int getLevel() {
		return level;
	}

	/**
	 * @return Returns the max.
	 */
	public int getMax() {
		return max;
	}

	/**
	 * @return Returns the min.
	 */
	public int getMin() {
		return min;
	}

	/**
	 * @return Returns the opposingcontacts.
	 */
	public List getOpposingContacts() {
		return opposingContacts;
	}

	/**
	 * @return Returns the price.
	 */
	public int getPrice() {
		return price;
	}

	/**
	 * @return Returns the terrible.
	 */
	public boolean isTerrible() {
		return terrible;
	}

	/**
	 * @return Returns the type.
	 */
	public char getType() {
		return type;
	}

	/**
	 * @return Returns the upgradable.
	 */
	public boolean isUpgradable() {
		return upgradable;
	}

	/**
	 * @return Returns the upgradePrice.
	 */
	public int getUpgradePrice() {
		return upgradePrice;
	}

	/**
	 * @return Returns the vertex.
	 */
	public Point[] getVertex() {
		return vertex;
	}

	/**
	 * @param group
	 *            The group to set.
	 */
	public void setCommand(Command command) {
		this.command = command;
	}

	/**
	 * @param lineNumber
	 *            The lineNumber to set.
	 */
	public void setLineNumber(int lineNumber) {
		this.lineNumber = lineNumber;
	}

	/**
	 * get step size
	 */
	public int getStep() {
		return getMovement();
	}

	/**
	 * @see fb.army.hierarchy.HierarchyNode#getNumberOfWarriors()
	 */
	public int getNumberOfWarriors() {
		return 0;
	}

	// --------------------------abstract methods-----------

	/**
	 * @return resource key
	 */
	protected abstract String getResourceKey();

	/**
	 * do move this entity to specified location
	 * 
	 * @param x
	 *            destination x coordinate
	 * @param y
	 *            destination y coordinate
	 * @param consumeMovement
	 *            must movement points be consumed
	 */
	public abstract void doMoveTo(int x, int y, boolean consumeMovement);

	/**
	 * do rotate this entity
	 * 
	 * @param theta
	 *            rotation angle
	 */
	public abstract void doRotate(double theta);

	/**
	 * remove this entity after killing
	 * 
	 */
	public abstract void commitKill();

	/**
	 * wound this entity
	 * 
	 * @param wounds
	 *            number of wounds
	 * @return remaining wounds
	 */
	public abstract int wound(int wounds);

	/**
	 * combat another entity
	 * 
	 * @param combatworth
	 *            number of dice
	 * @param target
	 *            ennemy target
	 * @return number of wounds
	 */
	public abstract int combat(int combatworth, Entity target);

	/**
	 * get front line width
	 * 
	 * @return front line width
	 */
	public abstract int getFrontWidth();

	/**
	 * rout this entity
	 * 
	 */
	public abstract void rout();

	/**
	 * remove this entity after routing
	 * 
	 */
	public abstract void commitRout();

	/**
	 * calculate contacts
	 * 
	 */
	public abstract void calculateContacts();

	/**
	 * clear contacts
	 * 
	 */
	public abstract void clearContacts();

	/**
	 * fight
	 * 
	 */
	public abstract void fight();

	/**
	 * get front leftmost warrior
	 * 
	 * @return front leftmost warrior
	 */
	public abstract Warrior leftModel();

	/**
	 * get front rightmost warrior
	 * 
	 * @return front rightmost warrior
	 */
	public abstract Warrior rightModel();

	/**
	 * teleport this entity to its destination
	 * 
	 */
	public abstract void teleport();

	/**
	 * kill entity warrior
	 * 
	 * @param model
	 *            the warrior
	 */
	public abstract void setDead(Warrior model);

	/**
	 * get max diagonal
	 * 
	 * @return maximum diagonal
	 */
	public abstract int getMaxDiag();

	/**
	 * calculate base surface
	 * 
	 * @return base surface
	 */
	public abstract float surface();

	/**
	 * calculate entity cost
	 */
	public abstract int getCost();

	/**
	 * paint movement range shape
	 * 
	 * @param g
	 *            graphics
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param scale
	 *            view scale
	 */
	public abstract void paintMovementRange(Graphics g, int dx, int dy,
			int scale);

	/**
	 * is this entity walking
	 * 
	 * @return true if this entity is walking
	 */
	public abstract boolean isWalking();

	/**
	 * is this entity dying
	 * 
	 * @return true if this entity is dying
	 */
	public abstract boolean isDying();

	/**
	 * stop entity action and make it stand
	 * 
	 */
	public abstract void stand();

	/**
	 * number of warriors for this entity
	 * 
	 * @return number of warriors
	 */
	public abstract int numberOfModels();

	/**
	 * start fight animation
	 * 
	 */
	public abstract void animFight();

	/**
	 * start threat animation
	 * 
	 */
	public abstract void animThreat();

	/**
	 * start death animation
	 * 
	 */
	public abstract void animKill();

	/**
	 * get x current coordinate
	 * 
	 * @return x current coordinate
	 */
	public abstract int getX();

	/**
	 * get y current coordinate
	 * 
	 * @return y current coordinate
	 */
	public abstract int getY();

	/**
	 * get entity total worth
	 * 
	 * @return entity total worth
	 */
	public abstract int totalWorth();

	/**
	 * change x current coordinate
	 * 
	 * @param x
	 *            new x coordinate
	 */
	public abstract void setX(int x);

	/**
	 * rotate entity to face another entity
	 * 
	 * @param e
	 *            the other entity
	 */
	public abstract void rotateTo(Entity e);

	/**
	 * change y current coordinate
	 * 
	 * @param y
	 *            new y coordinate
	 */
	public abstract void setY(int y);

	/**
	 * get entity total life points
	 * 
	 * @return entity total life points
	 */
	public abstract int getTotalLifePoints();

	/**
	 * get entity initial life points
	 * 
	 * @return entity initial life points
	 */
	public abstract int getInitialTotalLifePoints();

	/**
	 * heal this entity
	 * 
	 * @param lifePoints
	 *            added life points
	 */
	public abstract void heal(int lifePoints);

	/**
	 * get LeftMost Vertex
	 * 
	 * @return LeftMost Vertex
	 */
	public abstract Point getLeftMostVertex();

	/**
	 * get RightMost Vertex
	 * 
	 * @return RightMost Vertex
	 */
	public abstract Point getRightMostVertex();

	/**
	 * get Contact Area
	 * 
	 * @return Contact Area
	 */
	public abstract Area getContactArea();

	/**
	 * get Border
	 * 
	 * @return Border
	 */
	public abstract Shape getBorder();

	/**
	 * get Angle
	 * 
	 * @return Angle
	 */
	public abstract double getAngle();

	/**
	 * change angle
	 * 
	 * @param angle
	 *            new angle
	 */
	public abstract void setAngle(double angle);

	/**
	 * calculate distance to another entity
	 * 
	 * @param e
	 *            the other entity
	 * @return distance to other entity
	 */
	public abstract int distanceTo(Entity e);

	/**
	 * get translated line of movement
	 * 
	 * @param e
	 *            target entity
	 * @param tx
	 *            initial x coordinate
	 * @param ty
	 *            initial y coordinate
	 * @return line of movement length
	 */
	public abstract int translatedLineOfMovementTo(Entity e, int tx, int ty);

	/**
	 * get distance from a warrior
	 * 
	 * @param originmodel
	 *            the warrior
	 * @return distance from the warrior
	 */
	public abstract int distanceFrom(Warrior originmodel);

	/**
	 * get line of sight from a warrior
	 * 
	 * @param originmodel
	 *            the warrior
	 * @return line of sigh length
	 */
	public abstract int lineOfSightFrom(Warrior originmodel);

	/**
	 * get translated line of movement from a warrior
	 * 
	 * @param originmodel
	 *            the warrior
	 * @param tx
	 *            initial x coordinate
	 * @param ty
	 *            initial y coordinate
	 * @return line of sigh length
	 */
	public abstract int translatedLineOfMovementFrom(Warrior originmodel, int tx, int ty);

	/**
	 * create all entity warriors
	 * 
	 */
	public abstract void createModel();

	/**
	 * calculate base area
	 * 
	 */
	public abstract void calculateArea();

}