/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.artificialintelligence;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * implemention of A* algorythm to find the best path from one point to
 * another
 * 
 * @author farfpeuf
 * 
 */
public abstract class AStar  {

	/**
	 * origin and destination points
	 */
	protected Point origin, destination;

	/**
	 * explored nodes and nodes to explore
	 */
	protected List toExplore, explored;

	/**
	 * found path
	 */
	private LinkedList path;

	/**
	 * found path cost
	 */
	private int pathCost;

	/**
	 * current working node
	 */
	protected Node currentNode;

	/**
	 * stop the search ?
	 */
	private boolean stop = false;

	/**
	 * find point surrounding a location
	 * @param currentPosition the current point
	 * @return the surrounding points
	 */
	public abstract Point[] getSurroundingPoints(Point currentPosition);

	/**
	 * is this position legal
	 * @param position the position to test
	 * @return true if position is legal
	 */
	public abstract boolean isLegalPosition(Point position);

	/**
	 * calculate cost from parent node
	 * @param parent the parent position
	 * @param position the child position
	 * @return the cost
	 */
	public abstract int costFromParent(Point parent, Point position);

	/**
	 * estimate cost to destination
	 * @param position the current position
	 * @param destination the destination point
	 * @return the estimated cost
	 */
	public abstract int estimateCostToDestination(Point position,
			Point destination);

	/**
	 * constructor
	 * @param origin origin point
	 * @param destination destination point
	 */
	public AStar(Point origin, Point destination) {
		super();
		path = null;
		this.origin = origin;
		this.destination = destination;
		toExplore = new ArrayList();
		explored = new ArrayList();
	}

	/**
	 * find a node in a list specifying its position
	 * @param position the node position
	 * @param nodeList the node list
	 * @return the node or null if none
	 */
	public Node getNodeAt(Point position, List nodeList) {
		Iterator iter = nodeList.iterator();
		Node node;
		while (iter.hasNext()) {
			node = (Node) iter.next();
			if (node.position.equals(position)) {
				return node;
			}
		}
		return null;
	}

	/**
	 * stop the search
	 *
	 */
	public void stop() {
		stop = true;
	}

	/**
	 * run the search
	 *
	 */
	public void run() {

		Node firstNode=new Node(origin, null);
		toExplore.add(firstNode);
		currentNode=firstNode;
		int i;
		Point[] surroundingPoints;
		Node oldNode, newNode;
		int newCostFromOrig;

		Comparator nodeComparator = new Comparator() {

			public int compare(Object o1, Object o2) {
				Node node1 = (Node) o1;
				Node node2 = (Node) o2;
				if (node1.cost < node2.cost) {
					return -1;
				}
				if (node1.cost > node2.cost) {
					return 1;
				}
				return 0;
			}

		};

		while (!stopSearch()) {
			currentNode = (Node) toExplore.get(0);
			surroundingPoints = getSurroundingPoints(currentNode.position);
			toExplore.remove(currentNode);
			for (i = 0; i < surroundingPoints.length && !stop; i++) {
				if (getNodeAt(surroundingPoints[i], explored) == null
						&& isLegalPosition(surroundingPoints[i])) {
					oldNode = getNodeAt(surroundingPoints[i], toExplore);

					if (oldNode != null) {
						newCostFromOrig = currentNode.costFromOrig
								+ costFromParent(currentNode.position,
										surroundingPoints[i]);
						if (newCostFromOrig < oldNode.costFromOrig) {
							oldNode.parent = currentNode;
							oldNode.costFromOrig = newCostFromOrig;
							oldNode.cost = newCostFromOrig + oldNode.costToDest;
						}
					} else {
						newNode = new Node(surroundingPoints[i], currentNode);
						toExplore.add(newNode);
					}
				}

			}// for

			
			explored.add(currentNode);
			//FB.debug("x="+currentNode.position.x+" y="+currentNode.position.y+" cost="+currentNode.cost);
			if (!stop) {
				Collections.sort(toExplore, nodeComparator);
			}

		}// while

	}

	/**
	 * @return the found path as a point linked list
	 */
	public LinkedList getFoundPath() {
		if (path == null) {
			path = new LinkedList();
			pathCost = currentNode.cost;
			while (currentNode.parent != null) {
				path.addFirst(currentNode.position);
				currentNode = currentNode.parent;
			}
		}
		return path;
	}

	/**
	 * @return true if search must be stopped
	 */
	protected boolean stopSearch() {
		return (destination != null && currentNode.position.equals(destination))
				|| toExplore.isEmpty() || stop;
	}

	/**
	 * a path searching node
	 * @author farfpeuf
	 *
	 */
	public class Node {
		
		/**
		 * the node parent
		 */
		private Node parent;

		/**
		 * the node poistion
		 */
		private Point position;

		/**
		 * the node costs
		 */
		private int cost, costFromOrig, costToDest;

		/**
		 * constructor
		 * @param position the node position
		 * @param parent the node parent
		 */
		public Node(Point position, Node parent) {
			this.position = position;
			this.parent = parent;
			if (parent != null) {
				costFromOrig = parent.costFromOrig
						+ costFromParent(parent.position, position);
			} else {
				costFromOrig = 0;
			}
			if (destination != null) {
				costToDest = estimateCostToDestination(position, destination);
			} else {
				costToDest = 0;
			}
			cost = costFromOrig + costToDest;
		}

		/**
		 * @return Returns the costFromOrig.
		 */
		public int getCostFromOrig() {
			return costFromOrig;
		}

		/**
		 * @return Returns the position.
		 */
		public Point getPosition() {
			return position;
		}




	}

	/**
	 * @return Returns the pathCost.
	 */
	public int getPathCost() {
		return pathCost;
	}


}
