/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.campaign;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.Hashtable;

import fb.main.FB;
import fb.util.ArrayListModel;
import fb.util.FileTools;
/**
 * a hexagonal map data model
 * @author farfpeuf
 *
 */
public class HexagonalMapModel {
	
	//model file path
	private String path;

	//default tile type
	public static final char DEFAULT_TILE = 'a';

	//capital tile type
	public static final char CAPITAL_SET = 'K';

	//village tile type
	public static final char VILLAGE_SET = 'C';

	//used by property change listeners
	public static final String TILE_PROPERTY = "tile";

	//tiles table
	private char[][] tiles;

	//map size
	private int mapWidth = 10, mapLength = 10;

	//singleton instance
	private static HexagonalMapModel instance;

	//availables sides
	private ArrayListModel sides;

	//map side/capital
	private Hashtable capitals;

	/**
	 * get singleton instance
	 * @return the singleton
	 */
	public static synchronized HexagonalMapModel getInstance() {
		if (null == instance) {
			instance = new HexagonalMapModel();
		}
		return instance;
	}
/**
 * constructor
 * initializes lists maps and tables
 */
	private HexagonalMapModel() {
		sides = new ArrayListModel();
		capitals = new Hashtable();
		reset();
	}

	/**
	 * reset the data model
	 *
	 */
	public void reset() {
		tiles = new char[mapWidth][mapLength];
		capitals.clear();
		sides.clear();
	}

	/**
	 * load the model from a file
	 * @param path the file path
	 */
	public void loadFromFile(String path) {
		this.path = path;
		Reader reader = FileTools.getReader(path);
		load(reader);
	}

	/**
	 * load the model from a string
	 * @param mapSt the model as a String
	 */
	public void load(String mapSt) {
		StringReader reader = new StringReader(mapSt);
		load(reader);
	}

	/**
	 * load the model using a reader
	 * @param reader the reader
	 */
	public void load(Reader reader) {
		sides.clear();

		String line;
		BufferedReader buffer;
		try {

			buffer = new BufferedReader(reader);
			line = buffer.readLine();
			mapWidth = Integer.parseInt(line);
			line = buffer.readLine();
			mapLength = Integer.parseInt(line);
			line = buffer.readLine();
			int j = 0;
			tiles = new char[mapWidth][mapLength];
			char tile;
			while (line != null) {
				for (int i = 0; i < line.length(); i++) {
					tile = line.charAt(i);
					setTile(i, j, tile);
				}

				j++;
				line = buffer.readLine();
			}

			buffer.close();
			reader.close();
		} catch (Exception e) {
			FB.exception(e);
		}
	}

	/**
	 * change a tile type
	 * @param i row
	 * @param j column
	 * @param tile tile type
	 */
	public void setTile(int i, int j, char tile) {

		if (i >= 0 && j >= 0 && i < mapWidth && j < mapLength) {
			char oldTile = tiles[i][j];

			if (oldTile != tile) {
				String side;
				tiles[i][j] = tile;
				if (tile >= 'A' && tile <= 'F') {
					side = Character.toString(tile).toUpperCase();
					sides.add(side);
					capitals.put(side, new Point(i, j));
				}
			}

		}
	}

	/**
	 * test if a tile is a capital
	 * @param p the tile coordinates
	 * @return true if the tile is a capital
	 */
	public boolean isCapital(Point p) {
		if (getBounds().contains(p)) {
			char tile = tiles[p.x][p.y];
			return (tile >= 'A' && tile <= 'F');
		}
		return false;
	}

	/**
	 * test if a tile is a village
	 * @param p the tile coordinates
	 * @return true if the tile is a village
	 */
	public boolean isVillage(Point p) {
		if (getBounds().contains(p)) {
			return (tiles[p.x][p.y] == 'V');
		}
		return false;
	}

	/**
	 * test if a tile is a capital or a village
	 * @param p the tile coordinates
	 * @return true if the tile is a capital or a villaeg
	 */
	public boolean isVillageOrCapital(Point p) {
		if (getBounds().contains(p)) {
			char tile = tiles[p.x][p.y];
			return (tile >= 'A' && tile <= 'F') || (tile == 'V');
		}
		return false;
	}

	/**
	 * reload data model
	 *
	 */
	public void reload() {
		if (path != null) {
			loadFromFile(path);
		} else {
			reset();
		}
	}

	/**
	 * save data model to original file
	 *
	 */
	public void save() {
		if (path != null) {
			save(path);
		}
	}

	/**
	 * save data model to a file
	 * @param file the file path
	 */
	public void save(String file) {
		if (sides.size() < 2) {
			setTile(0, 0, 'A');
			if (sides.size() < 2) {
				setTile(mapWidth - 1, mapLength - 1, 'B');
			}
		}

		FileTools.writeFile(file, getModelString());
	}

	/**
	 * get the data model as a string (in order to save it to a file)
	 * @return the data model as a string
	 */
	public String getModelString() {
		StringBuffer result = new StringBuffer();
		result.append(mapWidth);
		result.append("\n");
		result.append(mapLength);
		result.append("\n");
		for (int j = 0; j < tiles.length; j++) {
			for (int i = 0; i < tiles[j].length; i++) {
				result.append(tiles[i][j]);
			}
			result.append("\n");
		}

		return result.toString();
	}



	/**
	 * @return Returns the mapLength.
	 */
	public synchronized int getMapLength() {
		return mapLength;
	}
	/**
	 * @return Returns the mapWidth.
	 */
	public synchronized int getMapWidth() {
		return mapWidth;
	}
	
	/**
	 * change map length
	 * @param mapLength new length
	 */
	public synchronized void setMapLength(int mapLength) {
		if (mapLength != this.mapLength && mapLength > 1) {
			this.mapLength = mapLength;
			char oldTiles[][] = tiles;
			tiles = new char[mapWidth][mapLength];
			for (int i = 0; i < Math.min(oldTiles.length, tiles.length); i++) {
				System.arraycopy(oldTiles[i], 0, tiles[i], 0, Math.min(
						oldTiles[i].length, tiles[i].length));
			}

		}
	}

	/**
	 * change map width
	 * @param mapWidth new map width
	 */
	public synchronized void setMapWidth(int mapWidth) {
		if (mapWidth != this.mapWidth && mapWidth > 1) {
			this.mapWidth = mapWidth;
			char oldTiles[][] = tiles;
			tiles = new char[mapWidth][mapLength];
			for (int i = 0; i < Math.min(oldTiles.length, tiles.length); i++) {
				System.arraycopy(oldTiles[i], 0, tiles[i], 0, Math.min(
						oldTiles[i].length, tiles[i].length));
			}

		}

	}

	/**
	 * get tile type
	 * @param i row
	 * @param j column
	 * @return tile type
	 */
	public synchronized char getTile(int i, int j) {
		if (i >= 0 && j >= 0 && i < mapWidth && j < mapLength) {
			char tile = tiles[i][j];
			if (tile != 0) {
				return tile;
			}
		}
		return DEFAULT_TILE;
	}

	/**
	 * get available sides
	 * @return sides list
	 */
	public ArrayListModel getSides() {
		return sides;
	}

	/**
	 * get capital position for specified side
	 * @param side the side
	 * @return the capital coordinates
	 */
	public Point getCapital(int side) {
		Point p = (Point) capitals.get("" + (char) ('A' + side));
		return p;
	}

	/**
	 * get map bounds
	 * @return  map bounds as a rectangle
	 */
	public Rectangle getBounds() {
		return new Rectangle(0, 0, mapWidth, mapLength);
	}

	/**
	 * @return Returns the path.
	 */
	public String getPath() {
		return path;
	}

}