/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.spells;

import fb.army.entities.Entity;
import fb.army.entities.SpellCaster;
import fb.main.FB;
import fb.network.Client;
import fb.network.MessagesConstants;
import fb.sound.SoundPlayer;
import fb.util.D6;

/**
 * mother class for all spells
 * 
 * @author farfpeuf
 * 
 */
public abstract class BasicSpell {
	/**
	 * spell name
	 */
	protected String name;

	/**
	 * spell caster
	 */
	protected SpellCaster caster;

	/**
	 * perform spell action
	 * 
	 */
	public void action() {
		if (castSpell()) {
			commitSpell();
		}

	}

	/**
	 * commit the spell effects
	 * 
	 */
	public void commitSpell() {
		if (FB.isNet()) {
			sendNetSpell();
		} else {
			doCommitSpell();
		}
	}

	/**
	 * send spell to remote computers
	 * 
	 */
	private void sendNetSpell() {
		Client.getInstance().sendAndWait(spellString());
	}

	/**
	 * @return string to send spell to remote computers
	 */
	protected String spellString() {
		return MessagesConstants.COMMITSPELL + name + "#" + caster.getId();
	}

	/**
	 * do commit the spell effects
	 * 
	 */
	public void doCommitSpell() {

		try {
			SoundPlayer.getInstance().play("magic");
			if (FB.getInstance().isGraphicMode()) {
				prepareEffects();
				caster.animSpell();
			}
			while (caster.isCastingSpell()) {
				Thread.sleep(125);

			}

			endSpell();
		} catch (InterruptedException e) {
			// FB.debug(name+" interrupted");
		}

	}

	/**
	 * prepare spell visual effects
	 * 
	 */
	public void prepareEffects() {
	}

	/**
	 * synchronize spell with remote computers
	 * 
	 * @throws InterruptedException
	 */
	protected void magicSynchronize() throws InterruptedException {
		if (FB.isNet()) {
			Client.getInstance().magicSynchronize();
		}
	}

	/**
	 * @return Returns the caster.
	 */
	public SpellCaster getCaster() {
		return caster;
	}

	/**
	 * @param caster
	 *            The caster to set.
	 */
	public void setCaster(SpellCaster caster) {
		this.caster = caster;
	}

	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}

	/**
	 * end the spell
	 * 
	 * @throws InterruptedException
	 */
	public abstract void endSpell() throws InterruptedException;

	/**
	 * cast the spell
	 * 
	 * @return true if the spell was successfull
	 */
	public abstract boolean castSpell();

	/**
	 * is this entity a potential target for specified actor
	 * 
	 * @param a
	 *            spell actor
	 * @param e
	 *            spell target
	 * @return true if this entity is a potential target for specified actor
	 */
	public abstract boolean isTarget(Entity a, Entity e);

	/**
	 * is this entity a potential target for current actor
	 * 
	 * @param e
	 *            spell target
	 * @return true if this entity is a potential target for current spell
	 *         caster
	 */
	public abstract boolean isTarget(Entity e);

	/**
	 * @return spell cost
	 */
	public abstract int cost();

	/**
	 * @return spell maximum cost
	 */
	public abstract int maxCost();

	/**
	 * @return spell minimum cost
	 */
	public abstract int minCost();

	/**
	 * @return spell average cost
	 */
	public abstract int averageCost();

	/**
	 * @return modifiers for artificial intelligence
	 */
	public int getWeightModifiers() {
		int poid = caster.getMagicPoints() / 20;
		int maxcost = maxCost();
		if (caster.getMagicPoints() < averageCost()) {
			poid -= 2;
		}
		if (caster.getMagicPoints() < maxcost) {
			poid -= 1;
		}
		if (caster.getTotalLifePoints() == 1) {
			poid += 1;
		}
		if (caster.getOpposingContacts().size() > 0) {
			poid += 1;
		}
		if (caster.getArmy().degradation() > 2) {
			poid += 2;
		}
		if (caster.getMagicPoints() > 2 * maxcost) {
			poid += 2;
		}
		if (caster.getMagicPoints() > 3 * maxcost) {
			poid += 1;
		}
		if (caster.getMagicPoints() > 4 * maxcost) {
			poid += 1;
		}
		return poid;
	}

	/**
	 * @return true if AI player should cast this spell
	 */
	public boolean shouldAICast() {
		if (caster.getMagicPoints() < minCost()) {
			return false;
		}
		int poid = 0;
		for (int i = 0; i < 3; i++) {
			poid = poid + D6.roll();
		}
		poid += getWeightModifiers();
		return poid > 18;
	}

}
