/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.spells;

import fb.army.entities.Entity;
import fb.army.entities.Heroic;
import fb.army.entities.Leader;
import fb.army.entities.SpellCaster;
import fb.util.D6;
import fb.util.Resources;

/**
 * Heals (restores lost life points) the friendly target
 * 
 * @author farfpeuf
 * 
 */
public class Heal extends TargetSpell implements ValuedSpell {
	/**
	 * number of life points to restore
	 */
	private int lifePoints = 1;

	/**
	 * @return string to send spell to remote computers
	 */
	protected String spellString() {
		return super.spellString() + "#" + currentValue();
	}

	/**
	 * constructor
	 * 
	 */
	public Heal() {
		super();
		name = "heal";
	}

	/**
	 * is this entity a potential target for specified actor
	 * 
	 * @param a
	 *            spell actor
	 * @param e
	 *            spell target
	 * @return true if this entity is a potential target for specified actor
	 */
	public boolean isTarget(Entity a, Entity e) {
		return !e.isEnnemy(a) && e.getInitialPdv() > 1 && e.hasLostLifePoints()
				&& a.lineOfSightTo(e) < 9999 && super.isTarget(a, e);
	}

	/**
	 * is this entity a potential target for current actor
	 * 
	 * @param e
	 *            spell target
	 * @return true if this entity is a potential target for current spell
	 *         caster
	 */
	public boolean isTarget(Entity e) {
		return e.getInitialPdv() > 1 && e.hasLostLifePoints()
				&& super.isTarget(e) && !e.isEnnemy(caster)
				&& caster.lineOfSightTo(e) < 9999;
	}

	/**
	 * cast the spell
	 * 
	 * @return true if the spell was successfull
	 */
	public boolean castSpell() {
		if (caster.spendMagicPoints(cost())) {
			return true;
		}
		return false;
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return spell cost
	 */
	public int cost(SpellCaster caster, Entity target) {
		int cost = (int) (0.999 + caster.lineOfSightTo(target) / 12)
				+ lifePoints;
		for (int i = 0; i < lifePoints; i++) {
			cost += D6.roll();
		}
		return cost;
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return spell maximum cost
	 */
	public int maxCost(SpellCaster caster, Entity target) {
		return lifePoints * 6 + (int) (caster.lineOfSightTo(target) / 12);
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return spell minimumu cost
	 */
	public int minCost(SpellCaster caster, Entity target) {
		return lifePoints * 1 + (int) (caster.lineOfSightTo(target) / 12);
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return average cost
	 */
	public int averageCost(SpellCaster caster, Entity target) {
		return (int) ((double) lifePoints * 3.5)
				+ (int) (caster.lineOfSightTo(target) / 12);
	}

	/**
	 * end the spell
	 * 
	 * @throws InterruptedException
	 */
	public void endSpell() throws InterruptedException {
		target.heal(lifePoints);
		magicSynchronize();
	}

	/**
	 * @return current spell value
	 */
	public int currentValue() {
		return lifePoints;
	}


	/**
	 * @return current spell value
	 */
	public int maxValue() {
		if (target != null) {
			return target.getInitialTotalLifePoints()
					- target.getTotalLifePoints();
		}
		return 10;
	}

	/**
	 * @return minimum value
	 */
	public int minValue() {
		return 1;
	}

	/**
	 * @return the value name
	 */
	public String valueName() {
		return Resources.getString("healValue");
	}

	/**
	 * set the value
	 * 
	 * @param value
	 *            the new value
	 */
	public void setValue(int value) {
		lifePoints = value;
	}

	/**
	 * @param potentialTarget
	 *            the target
	 * @return weight modifiers for artificial intelligence for specified target
	 */
	public int getWeight(Entity potentialTarget) {
		int weight = super.getWeight(potentialTarget);

		if (potentialTarget instanceof Heroic
				|| potentialTarget instanceof Leader
				|| potentialTarget instanceof SpellCaster) {
			weight += 3;
		}
		if (potentialTarget.getTotalLifePoints() == 1) {
			weight += 4;
		}
		return weight;
	}
}