/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import javax.swing.AbstractListModel;

/**
 * List and list model
 * @author frfpeuf
 *
 */
public class ArrayListModel extends AbstractListModel implements List {

    /**
     * the delegate list that contains the data
     */
    protected List delegate;

    /**
     * Default constructor. Uses an ArrayList for implementation.
     *  
     */
    public ArrayListModel() {
        super();
        delegate = new ArrayList();
    }

    public ArrayListModel(Object[] items) {
    	this(Arrays.asList(items));
    }
    
    
    /**
     * Uses an ArrayList for implementation.
     * 
     * @param collection
     *            the collection whose elements are to be placed into this model
     */
    public ArrayListModel(Collection collection) {
        super();
        delegate = new ArrayList(collection);
    }

    //-- methods for AbstractListModel implementation --

    /**
     * @see javax.swing.ListModel#getElementAt(int)
     */
    public Object getElementAt(int index) {
        return delegate.get(index);
    }

    /**
     * @see javax.swing.ListModel#getSize()
     */
    public int getSize() {
        return delegate.size();
    }

    //-- methods for List implementation --

    /**
     * @param delegateList
     */
    public ArrayListModel(List delegateList) {
        super();
        delegate = delegateList;
    }

    /**
     * @see List#add(int, java.lang.Object);
     */
    public void add(int index, Object element) {
        delegate.add(index, element);
        fireIntervalAdded(this, index, index);
    }

    /**
     * @see List#add(java.lang.Object);
     */
    public boolean add(Object o) {
        int index = delegate.size();
        boolean result = delegate.add(o);
        fireIntervalAdded(this, index, index);
        return result;
    }

    /**
     * @see List#addAll(int, java.util.Collection);
     */
    public boolean addAll(int index, Collection c) {
        boolean result = delegate.addAll(index, c);
        fireIntervalAdded(this, index, index + c.size() - 1);
        return result;
    }

    /**
     * @see List#addAll(java.util.Collection);
     */
    public boolean addAll(Collection c) {
        int index = delegate.size();
        boolean result = delegate.addAll(c);
        fireIntervalAdded(this, index, index + c.size() - 1);
        return result;
    }
  
    
    /**
     * @see List#clear();
     */
    public void clear() {
        int index1 = delegate.size() - 1;
        delegate.clear();
        if (index1 >= 0) {
            fireIntervalRemoved(this, 0, index1);
        }
    }

    /**
     * @see List#contains(java.lang.Object);
     */
    public boolean contains(Object o) {
        return delegate.contains(o);
    }

    /**
     * @see List#containsAll(java.util.Collection);
     */
    public boolean containsAll(Collection c) {
        return delegate.containsAll(c);
    }

    /**
     * @see List#get(int);
     */
    public Object get(int index) {
        return delegate.get(index);
    }

    /**
     * @see List#indexOf(java.lang.Object);
     */
    public int indexOf(Object o) {
        return delegate.indexOf(o);
    }

    /**
     * @see List#isEmpty();
     */
    public boolean isEmpty() {
        return delegate.isEmpty();
    }

    /**
     * @see List#iterator();
     */
    public Iterator iterator() {
        return delegate.iterator();
    }

    /**
     * @see List#lastIndexOf(java.lang.Object);
     */
    public int lastIndexOf(Object o) {
        return delegate.lastIndexOf(o);
    }

    /**
     * @see List#listIterator();
     */
    public ListIterator listIterator() {
        return delegate.listIterator();
    }

    /**
     * @see List#listIterator(int);
     */
    public ListIterator listIterator(int index) {
        return delegate.listIterator(index);
    }

    /**
     * @see List#remove(int);
     */
    public Object remove(int index) {
        Object result = delegate.get(index);
        delegate.remove(index);
        fireIntervalRemoved(this, index, index);
        return result;
    }

    /**
     * @see List#remove(java.lang.Object);
     */
    public boolean remove(Object o) {
        int index = indexOf(o);
        boolean result = delegate.remove(o);
        if (index >= 0) {
            fireIntervalRemoved(this, index, index);
        }
        return result;
    }



    /**
     * @see List#removeAll(java.util.Collection);
     */
    public boolean removeAll(Collection c) {

        List intervals = new ArrayList();//used to memorise the deleted
                                         // intervals
        int index1;
        index1 = -1;
        int size = delegate.size();
        for (int i = 0; i < size; i++) {
            if (c.contains(delegate.get(i))) {
                if (index1 == -1) {//open interval
                    index1 = i;
                }
            } else {
                if (index1 != -1) {//close interval
                    intervals.add(new Interval(index1, i - 1));
                    index1 = -1;
                }
            }
        }

        //close last interval
        if (index1 != -1) {
            intervals.add(new Interval(index1, size - 1));
            index1 = -1;
        }

        boolean modified = delegate.removeAll(c);

        Iterator iter = intervals.iterator();
        Interval inter;
        while (iter.hasNext()) {
            inter = (Interval) iter.next();
            fireIntervalRemoved(this, inter.getIndex1(), inter.getIndex2());
        }

        return modified;
    }

    /**
     * @see List#retainAll(java.util.Collection);
     */
    public boolean retainAll(Collection c) {
        List intervals = new ArrayList();//used to memorise the deleted
                                         // intervals
        int index1;
        index1 = -1;
        int size = delegate.size();
        for (int i = 0; i < size; i++) {
            if (!c.contains(delegate.get(i))) {
                if (index1 == -1) {//open interval
                    index1 = i;
                }
            } else {
                if (index1 != -1) {//close interval
                    intervals.add(new Interval(index1, i - 1));
                    index1 = -1;
                }
            }
        }

        //close last interval
        if (index1 != -1) {
            intervals.add(new Interval(index1, size - 1));
            index1 = -1;
        }
        boolean modified = delegate.retainAll(c);

        Iterator iter = intervals.iterator();
        Interval inter;
        while (iter.hasNext()) {
            inter = (Interval) iter.next();
            fireIntervalRemoved(this, inter.getIndex1(), inter.getIndex2());
        }

        return modified;
    }

    /**
     * @see List#set(int, java.lang.Object);
     */
    public Object set(int index, Object element) {
        Object result = delegate.get(index);
        delegate.set(index, element);
        fireContentsChanged(this, index, index);
        return result;
    }

    /**
     * @see List#size();
     */
    public int size() {
        return delegate.size();
    }

    /**
     * @see List#subList(int, int);
     */
    public List subList(int fromIndex, int toIndex) {
        return delegate.subList(fromIndex, toIndex);
    }

    /**
     * @see List#toArray();
     */
    public Object[] toArray() {
        return delegate.toArray();
    }

    /**
     * @see List#toArray(java.lang.Object[]);
     */
    public Object[] toArray(Object[] a) {
        return delegate.toArray(a);
    }
    public void fireContentsChanged(Object source, int index0, int index1) {
        super.fireContentsChanged(source,index0,  index1);
    }
    //a class to store the intervals for the removeAll and retainAll methods
    private class Interval {
        private int index1, index2;

        /**
         * @param index1
         * @param index2
         */
        public Interval(int index1, int index2) {
            super();
            this.index1 = index1;
            this.index2 = index2;
        }

        /**
         * Get the index1.
         * 
         * @return Returns the index1.
         */
        public int getIndex1() {
            return this.index1;
        }

        /**
         * Get the index2.
         * 
         * @return Returns the index2.
         */
        public int getIndex2() {
            return this.index2;
        }
    }

}