/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.util;

import java.awt.Point;
import java.awt.Polygon;
import java.awt.geom.Area;

import fb.army.entities.Unit;
import fb.battlefieldelements.Warrior;

/**
 * utility class to calculate unit vertice and areas
 * 
 * @author farfpeuf
 * 
 */
public class UnitTools {

	/**
	 * calculate unit vertice
	 * 
	 * @param unit
	 *            the unit
	 * @return the vertice
	 */
	public static Point[] calculateVertex(Unit unit) {
		return calculateVertex(unit.getFront(), unit.getNumber(), unit
				.getOrientation(), unit.getX(), unit.getY(), unit.getWidth(),
				unit.rowLength());
	}

	/**
	 * calculate unit vertice
	 * 
	 * @param front
	 *            number of warriors at the front line
	 * @param number
	 *            number of warriors
	 * @param orientation
	 *            unit orientation angle
	 * @param x
	 *            unit x coordinate
	 * @param y
	 *            unit y coordinate
	 * @param width
	 *            warriors base width
	 * @param length
	 *            warriors base length
	 * @return the vertice
	 */
	public static Point[] calculateVertex(int front, int number,
			double orientation, int x, int y, float width, float length) {

		int reste;
		int ax, ay, w1, l1, gx, gy, w2, l2;
		Point[] result;

		reste = number % front;

		ax = x - (int) (((float) front) / 2 * width * 8);
		if (front % 2 == 0) {
			ax -= (int) (width * 8 / 2);
		}
		ay = y - (int) (length * 8 / 2);

		w1 = (int) (front * width * 8);
		l1 = (int) (((int) (number / front)) * length * 8);

		if (reste > 0) {

			gx = x - (int) (((float) reste) / 2 * width * 8);
			if (front % 2 == 0) {
				gx -= (int) (width * 8 / 2);
			}
			gy = ay + l1;

			w2 = (int) (reste * width * 8);
			l2 = (int) (length * 8);

			result = new Point[8];
			// a
			result[0] = new Point(ax, ay);
			// b
			result[1] = new Point(ax + w1, ay);
			// c
			result[2] = new Point(ax + w1, ay + l1);
			// d
			result[3] = new Point(gx + w2, gy);
			// e
			result[4] = new Point(gx + w2, gy + l2);
			// f
			result[5] = new Point(gx, gy + l2);
			// g
			result[6] = new Point(gx, gy);
			// h
			result[7] = new Point(ax, ay + l1);

		} else {
			result = new Point[4];

			// a
			result[0] = new Point(ax, ay);
			// b
			result[1] = new Point(ax + w1, ay);
			// c
			result[2] = new Point(ax + w1, ay + l1);
			// h
			result[3] = new Point(ax, ay + l1);

		}

		rotateVertex(result, orientation, x, y);
		return result;

	}

	/**
	 * rotate vertice
	 * 
	 * @param vertex
	 *            the vertice
	 * @param theta
	 *            rotation angle
	 * @param xc
	 *            rotation center x coordinate
	 * @param yc
	 *            rotation center y coordinate
	 */
	public static void rotateVertex(Point[] vertex, double theta, int xc, int yc) {

		for (int i = 0; i < vertex.length; i++) {
			GeometryTools.rotatePoint(theta, vertex[i], xc, yc);
		}

	}

	/**
	 * translate vertice
	 * 
	 * @param vertex
	 *            the vertice
	 * @param xc
	 *            translation x offset
	 * @param yc
	 *            translation y offset
	 */
	public static void translateVertex(Point[] vertex, int dx, int dy) {

		for (int i = 0; i < vertex.length; i++) {
			vertex[i].translate(dx, dy);
		}

	}

	/**
	 * calculate left 90 manoeuvre path area
	 * 
	 * @param front
	 *            number of warriors at the front line
	 * @param number
	 *            number of warriors
	 * @param orientation
	 *            unit orientation angle
	 * @param x
	 *            unit x coordinate
	 * @param y
	 *            unit y coordinate
	 * @param width
	 *            warriors base width
	 * @param length
	 *            warriors base length
	 * @param nbRows
	 *            unit number of rows
	 * @param vertex
	 *            unit vertice
	 * @return the manoeuvre path area
	 */
	public static Area left90PathArea(int front, int number,
			double orientation, int x, int y, float width, float length,
			int nbRows, Point[] vertex) {

		int nx, ny;

		nx = (int) ((-(float) (front - 1) / 2) * width * 8 + x);
		if (front % 2 == 0) {
			nx -= (int) (width * 8 / 2);
		}
		ny = (int) ((nbRows - 1) * length * 8 / 2 + y);
		if (nbRows % 2 == 0) {
			ny -= (int) (length * 8 / 2);
		}

		Point p = new Point(nx, ny);
		GeometryTools.rotatePoint(orientation, p, x, y);

		Point[] result = calculateVertex(nbRows, number, orientation
				+ java.lang.Math.PI / 2, (int) p.getX(), (int) p.getY(), width,
				length);
		return calculatePathArea(vertex, result, x, y, (int) p.getX(), (int) p
				.getY());

	}

	/**
	 * calculate right 90 manoeuvre path area
	 * 
	 * @param front
	 *            number of warriors at the front line
	 * @param number
	 *            number of warriors
	 * @param orientation
	 *            unit orientation angle
	 * @param x
	 *            unit x coordinate
	 * @param y
	 *            unit y coordinate
	 * @param width
	 *            warriors base width
	 * @param length
	 *            warriors base length
	 * @param nbRows
	 *            unit number of rows
	 * @param vertex
	 *            unit vertice
	 * @return the manoeuvre path area
	 */
	public static Area right90PathArea(int front, int number,
			double orientation, int x, int y, float width, float length,
			int nbRows, Point[] vertex) {

		int nx, ny;

		nx = (int) (((float) (front - 1) / 2) * width * 8 + x);
		if (front % 2 == 0) {
			nx -= (int) (width * 8 / 2);
		}
		ny = (int) ((nbRows - 1) * length * 8 / 2 + y);
		if (nbRows % 2 == 0) {
			ny += (int) (length * 8 / 2);
		}

		Point p = new Point(nx, ny);
		GeometryTools.rotatePoint(orientation, p, x, y);

		Point[] result = calculateVertex(nbRows, number, orientation
				- java.lang.Math.PI / 2, (int) p.getX(), (int) p.getY(), width,
				length);
		return calculatePathArea(vertex, result, x, y, (int) p.getX(), (int) p
				.getY());

	}

	/**
	 * calculate wheel manoeuvre path area
	 * 
	 * @param vertex
	 *            unit vertice
	 * @param x
	 *            unit x coordinate
	 * @param y
	 *            unit y coordinate
	 * @param theta
	 *            wheel angle
	 * @param model
	 *            model around which wheel is performed
	 * @return the manoeuvre path area
	 */

	public static Area wheelPathArea(Point[] vertex, int x, int y,
			double theta, Warrior model) {

		Point[] result = new Point[vertex.length];
		for (int i = 0; i < vertex.length; i++) {
			result[i] = (Point) vertex[i].clone();
		}

		rotateVertex(result, theta, model.getX(), model.getY());
		return calculatePathArea(vertex, result, x, y, model.getX(), model
				.getY());

	}

	/**
	 * calculate manoeuvre path area
	 * 
	 * @param overtex
	 *            old unit vertice
	 * @param nvertex
	 *            new unit vertice
	 * @param ox
	 *            old unit x coordinate
	 * @param oy
	 *            old unit y coordinate
	 * @param nx
	 *            new unit x coordinate
	 * @param ny
	 *            new unit y coordinate
	 * @return the manoeuvre path area
	 */
	public static Area calculatePathArea(Point[] overtex, Point[] nvertex,
			int ox, int oy, int nx, int ny) {
		Area parcours = new Area();
		int[] areax, areay;
		areax = new int[4];
		areay = new int[4];
		parcours.add(vertexToArea(nvertex));

		if (overtex.length == nvertex.length) {

			areax[0] = ox;
			areax[1] = nx;
			areay[0] = oy;
			areay[1] = ny;

			for (int i = 0; i < overtex.length; i++) {

				areax[2] = (int) overtex[i].getX();
				areax[3] = (int) nvertex[i].getX();

				areay[2] = (int) overtex[i].getY();
				areay[3] = (int) nvertex[i].getY();

				parcours.add(new Area(new Polygon(areax, areay, 4)));

			}
		}
		return parcours;
	}

	/**
	 * calculate area of polygon defined by vertice
	 * 
	 * @param vertex
	 *            the vertice
	 * @return the area
	 */
	public static Area vertexToArea(Point[] vertex) {
		Polygon p = new Polygon();
		for (int i = 0; i < vertex.length; i++) {
			p.addPoint((int) vertex[i].getX(), (int) vertex[i].getY());
		}
		return new Area(p);
	}

}